import math
import threading

from collections import deque
from itertools import chain

from cellsepi.backend.main_window.expert_mode.event_manager import EventManager
from cellsepi.backend.main_window.expert_mode.listener import ErrorEvent, OnPipelineChangeEvent, ModuleExecutedEvent, \
    ModuleStartedEvent, PipelinePauseEvent, PipelineCancelEvent, PipelineErrorEvent
from cellsepi.backend.main_window.expert_mode.module import Module,Port
from cellsepi.backend.main_window.expert_mode.pipe import Pipe
from typing import List, Dict, Type


class Pipeline:
    def __init__(self):
        self.modules: List[Module] = [] #running order
        self.module_map: Dict[str, Module] = {} #mapping for fast access to the modules
        self.pipes_in: Dict[str,List[Pipe]] = {} #dict[target,[Pipe]]
        self.pipes_out: Dict[str,List[Pipe]] = {} #dict[source,[Pipe]]
        self.run_order: deque[str] = deque()
        self.modules_executed = 0
        self.executing: str = ""
        self.running: bool = False
        self._continue_event = threading.Event()
        self._cancel_event = threading.Event()
        self.event_manager: EventManager = EventManager()

    def add_module(self, module_class: Type[Module]) -> Module:
        """
        Creates a module of the given class and adds it to the pipeline.
        """
        module = module_class(module_id=module_class.get_id())
        return self._add_module(module,module_class)

    def add_module_with_id(self, module_class: Type[Module],module_id: str) -> Module:
        """
        Creates a module of the given class with the given module_id and adds it to the pipeline.
        """
        if not module_id.startswith(module_class.gui_config().name):
            raise ValueError(f"Invalid module id '{module_id}' for the module_class '{module_class.gui_config().name}'")
        module = module_class(module_id=module_id)
        module.occupy()
        return self._add_module(module,module_class)

    def _add_module(self, module: Module,module_class: Type[Module]) -> Module:
        self.modules.append(module)
        self.module_map[module.module_id] = module
        module.event_manager = self.event_manager
        self.pipes_in[module.module_id] = []
        self.pipes_out[module.module_id] = []
        self.event_manager.notify(OnPipelineChangeEvent(f"Added module {module_class.gui_config().name}"))
        return module

    def get_new_module_id(self, module_id_old: str):
        """
        Gets a new module id for the module with the given module_id_old.
        """
        if not module_id_old in self.module_map:
            raise ValueError(f"Module id '{module_id_old}' doesen't exists")
        self.module_map[module_id_old].free_id_number(self.module_map[module_id_old].get_id_number())
        module_id_new = self.module_map[module_id_old].get_id()
        self.module_map[module_id_old].module_id = module_id_new
        self.module_map[module_id_new] = self.module_map[module_id_old]
        del self.module_map[module_id_old]
        self.pipes_in[module_id_new] = self.pipes_in[module_id_old]
        del self.pipes_in[module_id_old]
        self.pipes_out[module_id_new] = self.pipes_out[module_id_old]
        del self.pipes_out[module_id_old]

    def remove_module(self, module: Module) -> None:
        """
        Removes a module from the pipeline.
        Raises:
            RuntimeError: If the module is still connected to other modules.
        """
        if not self.is_disconnected(module.module_id):
            raise RuntimeError(f"Cannot remove module '{module.module_id}' from pipeline while connections to other modules still exists.")
        if module in self.modules:
            self.modules.remove(module)
            del self.module_map[module.module_id]
            del self.pipes_in[module.module_id]
            del self.pipes_out[module.module_id]
            module.destroy()
            self.event_manager.notify(OnPipelineChangeEvent(f"Removed module {module.gui_config().name}"))

    def is_disconnected(self, module_name: str) -> bool:
        """
        Checks if a module has no connected modules.
        """
        return len(self.pipes_in[module_name]) == 0 and len(self.pipes_out[module_name]) == 0

    def remove_connection(self,source_id: str, target_id: str) -> None:
        """
        Removes a pipe between the source and target modules.
        Raises:
            ValueError: If a pipe between the source and target modules does not exist.
        """
        pipe = self.get_pipe(source_id, target_id)
        if pipe is None:
            raise ValueError(f"Pipe between source module '{source_id}' and target module '{target_id}' does not exist.")

        for port in pipe.ports:
            pipe.target_module.inputs[port].data = None

        self.pipes_in[target_id].remove(pipe)
        self.pipes_out[source_id].remove(pipe)
        self.event_manager.notify(OnPipelineChangeEvent(f"Removed connection between {target_id} and {source_id}"))

    def get_pipe(self, source_name: str, target_name: str) -> Pipe | None:
        """
        Returns the pipe between source module and target module or if it does not exist it returns None.
        """
        for pipe in self.pipes_in.get(target_name, []):
            if pipe.source_module.module_id == source_name:
                return pipe
        return None

    def add_connection(self, pipe: Pipe) -> None:
        """
        Adds a pipe to the pipeline.
        Raises:
            ModuleNotFoundError: if the target or source module is not already added in the pipeline.
            ValueError: If a pipe between the target and source module exists already in the pipeline.
        """
        if pipe.source_module not in self.modules:
            raise ModuleNotFoundError(f"Source module '{pipe.source_module.module_id}' not found in the pipeline.")
        if pipe.target_module not in self.modules:
            raise ModuleNotFoundError(f"Target module '{pipe.target_module.module_id}' not found in the pipeline.")

        if self.check_connections(pipe.source_module.module_id, pipe.target_module.module_id) is not None:
                raise ValueError(f"Pipe between source module '{pipe.source_module.module_id}' and target module '{pipe.target_module.module_id}' already exists.")

        self.pipes_in[pipe.target_module.module_id].append(pipe)
        self.pipes_out[pipe.source_module.module_id].append(pipe)

        self.event_manager.notify(OnPipelineChangeEvent(f"Added connection between {pipe.target_module.module_id} and {pipe.source_module.module_id}"))

    def expand_connection(self,pipe:Pipe, ports: List[str]) -> None:
        """
        Expands the ports tranfered with the pipe between the source and target modules with the given ports.
        """
        pipe.ports.extend(ports)
        self.event_manager.notify(OnPipelineChangeEvent(
            f"Expanded the connection between {pipe.target_module.module_id} and {pipe.source_module.module_id}"))

    def check_connections(self,source_module_id:str,target_module_id:str) -> Pipe | None:
        """
        Checks if a pipe between the source and target modules exist in the pipeline.
        """
        for existing_pipe in self.pipes_in[target_module_id]:
            if existing_pipe.source_module.module_id == source_module_id:
                return existing_pipe
        return None

    def check_ports_occupied(self,module_id: str,ports:List[str]) -> bool:
        """
        Checks if the given module ports are occupied by existing pipes.
        """
        for port in ports:
            for pipe in self.pipes_in[module_id]:
                if port in pipe.ports:
                    return True
        return False

    def setup_incoming_degree(self) -> Dict[str, int]:
        """
        Returns a dictionary mapping module names to incoming degree (incoming degree is how many pipes are going into a module).
        """
        return {module.module_id: len(self.pipes_in[module.module_id]) for module in self.modules}

    def get_run_order(self) -> deque[str]:
        """
        Get the topologic orders with Kahn's algorithm.
        For this, the algorithm uses the graph given by the pipes of the pipeline.
        """
        topological_order: deque[str] = deque()
        in_degree = self.setup_incoming_degree()
        queue = deque()
        for module_name, degree in in_degree.items():
            if degree == 0:
                queue.append(module_name)
        while queue:
            module_name = queue.popleft()
            topological_order.append(module_name)
            del in_degree[module_name]
            for pipe in self.pipes_out[module_name]:
                in_degree[pipe.target_module.module_id] -= 1
                if in_degree[pipe.target_module.module_id] == 0:
                    queue.append(pipe.target_module.module_id)
        if len(topological_order) != len(self.modules):
            raise RuntimeError(f"The pipeline contains a cycle, only acyclic graphs are supported.")
        return topological_order

    def check_pipeline_runnable(self,ignore:List[str]="") -> bool:
        """
        Checks if every module input is satisfied.
        """
        for module in self.modules:
            if module.module_id in ignore:
                continue
            if not self.check_module_satisfied(module.module_id):
                return False
        return True

    def check_module_satisfied(self,module_id: str) -> bool:
        """
        Checks if a modules inputs are satisfied.
        """
        module_pipes = self.pipes_in[module_id]
        delivered_ports = set(chain.from_iterable(pipe.ports for pipe in module_pipes))
        if not all(port_name in delivered_ports for port_name in self.module_map[module_id].get_mandatory_inputs()):
            return False
        else:
            return True

    def check_module_runnable(self,module_name: str) -> bool:
        """
        Checks if the module input port data from the given module_name is not None.
        """
        if not all(self.module_map[module_name].inputs[port_name].data is not None for port_name in self.module_map[module_name].get_mandatory_inputs()):
            return False
        else:
            return True

    def run(self,ignore_modules: List[str] = None) -> None:
        """
        Executes the steps of the Pipeline.
        Skips steps of the Pipeline if min. one of the mandatory inputs is None.

        Arguments:
            ignore_modules: List of modules to ignore.
        """
        self.modules_executed = 0
        self._continue_event.clear()
        try:
            self.run_order = self.get_run_order()
        except RuntimeError as e:
            self.event_manager.notify(PipelineErrorEvent("Cycle in Pipeline",e.args[0]))
            return
        while self.run_order:
            self.running = True
            module_name = self.run_order.popleft()
            if ignore_modules is not None and module_name in ignore_modules:
                continue
            module = self.module_map[module_name]
            module_pipes = self.pipes_in[module.module_id]
            for pipe in module_pipes:
                pipe.run()
            if self.check_module_runnable(module_name):
                try:
                    self.executing = module_name
                    self.event_manager.notify(ModuleStartedEvent(module_name))
                    pause = module.run() #if the run of a module returns True, the module wants to stop the pipeline.
                    if pause and not self._cancel_event.is_set():
                        self.event_manager.notify(PipelinePauseEvent(module_name))
                        self._continue_event.wait()
                        self.event_manager.notify(PipelinePauseEvent(module_name,True))
                        self._continue_event.clear()
                    module.finished()
                    self.executing = ""
                    self.modules_executed += 1
                    self.event_manager.notify(ModuleExecutedEvent(module_name))
                    if self._cancel_event.is_set():
                        self.running = False
                        self.event_manager.notify(PipelineCancelEvent(self.executing))
                        self._cancel_event.clear()
                        return

                except PipelineRunningException as e:
                    self.running = False
                    self.event_manager.notify(ErrorEvent(e.error_type,e.description))
                    self.executing = ""
                    self._cancel_event.clear()
                    return
            else:
                self.modules_executed += 1
                self.event_manager.notify(ModuleExecutedEvent(module_name))
                continue
        self.running = False

    def resume(self):
        """
        Resumes the pipeline after the execution of the pipeline got paused.
        """
        self._continue_event.set()

    def cancel(self):
        """
        Cancels the pipeline after the currently executed module is finished.
        """
        self._cancel_event.set()
        self._continue_event.set()

class PipelineRunningException(Exception):
    """
    Exception raised if a module in the pipeline has an error and the pipeline needs to stop.
    """
    def __init__(self, error_type: str, description: str):
        self.error_type = error_type
        self.description = description
        super().__init__(f"{error_type}: {description}")


from concurrent.futures import ThreadPoolExecutor
import numpy as np


def calculate_mask_diameters(mask):
    """
    Calculates the diameter of each cell in the given mask by assuming circular cells.

    Attributes:
        mask: the mask of an image
    Returns:
        list[]: the diameter of each cell in the given mask
    """
    cell_ids = np.unique(mask)
    diameters = []
    cell_ids = cell_ids[1:]

    for cell_id in cell_ids:
        cell_mask = mask == cell_id
        cell_area = np.sum(cell_mask)

        # Approximate diameter assuming circular cells
        cell_diameter = 2 * np.sqrt(cell_area / np.pi)
        diameters.append(cell_diameter)

    return diameters

class AverageDiameter:
    """
    This class calculates the average diameter of a cell in all the given images.
    """
    def __init__(self, gui):
        self.gui = gui
        self.csp = gui.csp

    def get_avg_diameter(self):
        """
        This method calculates the average diameter of a cell in all the given images.
        Returns: The average diameter of all cells rounded to 2 decimal places.
        """
        mask_paths = self.csp.mask_paths
        if mask_paths is not None:
            segmentation_channel = self.csp.config.get_bf_channel()
            all_diameters = []
            valid_image_id = [
                key for key in mask_paths.keys()
                if isinstance(mask_paths[key], dict) and segmentation_channel in mask_paths[key]
            ]

            def process_image(image_id):
                mask_path = mask_paths[image_id][segmentation_channel]
                mask_data = np.load(mask_path, allow_pickle=True).item()
                mask = mask_data["masks"]
                return calculate_mask_diameters(mask)
            try:
                with ThreadPoolExecutor() as executor:
                    results = executor.map(process_image, valid_image_id)

                for diameters in results:
                    all_diameters.extend(diameters)

                if len(all_diameters) == 0:
                    return 0.00

                rounded_diameters = round(np.mean(all_diameters), 2)
                self.gui.training_environment.diameter = rounded_diameters
                if self.gui.training_environment.field_diameter.disabled is False:
                    self.gui.training_environment.field_diameter.value = rounded_diameters
                    self.gui.training_environment.field_diameter.update()
                return rounded_diameters
            except:
                return self.gui.diameter_text.value





import base64
from io import BytesIO
from pathlib import Path

from cellsepi.backend.main_window.cellsepi import CellSePi
import numpy as np
from PIL import Image
from collections import defaultdict


class Mask:
    """
    in the class the created numpy files of the mask are converted
    in a displayable format (png)

    Attributes:
        csp=current CellSePi object
        mask_outputs= stores the already converted mask outputs, consists
                      of image_id and the path
    """
    def __init__(self,csp:CellSePi):
        self.csp= csp
        # the path to the already generated masks are stored in here
        self.mask_outputs = defaultdict(dict)# [image_id,path zu .png]

    def load_mask_into_canvas(self):
        """
        loads the numpy files of the mask to the id and converts it to png
        """
        #iterate over the processed data to load the mask images for the current image
        image_id = self.csp.image_id
        bfc= self.csp.config.get_bf_channel()
        if image_id in self.csp.mask_paths:

            #load the npy file and convert it to directory
            mask_data = np.load(Path(self.csp.mask_paths[image_id][bfc]), allow_pickle=True).item()

            #extract the mask data and the outline of the cell
            mask= mask_data["masks"]
            outline = mask_data["outlines"]
            self.convert_npy_to_canvas(mask,outline)



    def convert_npy_to_canvas(self,mask, outline):
        """
        handles the conversion of the given file data

        Args:
            mask= the mask data stored in the numpy directory
            outline= the outline data stored in the numpy directory
        """
        buffer= BytesIO()
        if mask.ndim == 3:
            mask = np.max(mask, axis=0)
        if outline.ndim == 3:
            outline = np.max(outline, axis=0)

        image_mask = np.zeros(shape=(mask.shape[0], mask.shape[1], 4), dtype=np.uint8)
        r,g,b = self.csp.config.get_mask_color()
        image_mask[mask != 0] = (r, g, b, self.csp.color_opacity)
        r, g, b = self.csp.config.get_outline_color()
        image_mask[outline != 0] = (r, g, b, 255)
        im= Image.fromarray(image_mask).convert("RGBA")
        im.resize(size=(700,500))

        #saves the image as a image(base64)
        im.save(buffer, format="PNG")
        buffer.seek(0)
        image_base_64= base64.b64encode(buffer.getvalue()).decode('utf-8')

        #create the output path independent of operating system
        brightfield_channel=self.csp.config.get_bf_channel()
        image_id=self.csp.image_id

        #saves the created output image.
        self.mask_outputs[image_id][brightfield_channel]=image_base_64










import asyncio
import base64
import os
from io import BytesIO
import flet as ft
import cv2
import numpy as np
from PIL import Image, ImageEnhance
from matplotlib import pyplot as plt
from tifffile import tifffile

from cellsepi.frontend.main_window.gui_canvas import update_main_image


class ImageTuning:
    """
    Manages the task to tune the image brightness and contrast.
    Attributes:
        gui (GUI): The GUI object that contains all objects for the gui.
        running_tasks: contains all task that are currently running.
        cached_image: contains the image that is current be cached.
    """
    def __init__(self,gui):
        self.gui = gui
        self.running_tasks = set()
        self.cached_image = None
    async def update_brightness_and_contrast_async(self, on_click=False,linux_or_3d = False):
        """
        Updates the main image brightness and contrast with the current selected values with the sliders.

        If a new image was clicked all old tasked got cancel so the new image has prio over the old one.

        Args:
            on_click (bool): if a new main image is clicked or not.
            linux_or_3d (bool): if the program is running on Linux or the images are 3d.
        """
        if linux_or_3d and on_click:
            self.cancel_all_tasks()
            self.gui.canvas.main_image.content.src_base64 = self.gui.csp.linux_images[self.gui.csp.image_id][self.gui.csp.channel_id]
            self.gui.canvas.main_image.update()
        elif on_click:
            self.cancel_all_tasks()
            self.gui.canvas.main_image.content.src_base64 = None
            self.gui.canvas.main_image.content.src = self.gui.csp.image_paths[self.gui.csp.image_id][self.gui.csp.channel_id]
            self.gui.canvas.main_image.update()
        else:
            task = asyncio.create_task(self.update_image())
            self.running_tasks.add(task)
            try:
                await task
            except asyncio.CancelledError:
                pass
            finally:
                self.running_tasks.discard(task)

    def cancel_all_tasks(self):
        for task in self.running_tasks:
            task.cancel()
        self.running_tasks.clear()


    async def update_image(self):
        """
        Updates the main image as base64_image with the new brightness and contrast values.
        """
        base64_image = await self.adjust_image_async(
            round(self.gui.brightness_slider.value, 2),
            round(self.gui.contrast_slider.value, 2)
        )
        self.gui.canvas.main_image.content.src_base64 = base64_image
        self.gui.canvas.main_image.update()

    async def adjust_image_async(self, brightness, contrast):
        return await asyncio.to_thread(self.adjust_image_in_memory, brightness, contrast)

    def adjust_image_in_memory(self, brightness, contrast):
        """
        Gets the current selected image and changes its brightness and contrast.
        Without saving it to disk.

        Args:
            brightness (int): the selected image's brightness'.
            contrast (int): the selected image's contrast'.

        Returns:
            image (base64): the updated image.
        """
        image_path = self.gui.csp.image_paths[self.gui.csp.image_id][self.gui.csp.channel_id]
        image = self.load_image(image_path)

        enhancer = ImageEnhance.Brightness(image)
        image = enhancer.enhance(brightness)

        enhancer = ImageEnhance.Contrast(image)
        image = enhancer.enhance(contrast)

        buffer = BytesIO()
        image.save(buffer, format="PNG")
        buffer.seek(0)

        return base64.b64encode(buffer.getvalue()).decode('utf-8')

    def load_image(self, image_path):
        """
        Checks if the image is already loaded.
        If not, it gets the new one from the image_path

        Args:
            image_path (str): the path to the image

        Returns:
            image (pillow): the loaded image.
        """
        if self.cached_image and self.cached_image[0] == image_path:
            return self.cached_image[1]
        image = tifffile.imread(image_path)
        if image.ndim == 3:
            image = np.max(image, axis=2)
        img = Image.fromarray(image)
        self.cached_image = (image_path, img)
        return img

    def save_current_main_image(self):
        """
        Saves the current selected image to disk.
        """
        if self.gui.csp.adjusted_image_path is None:
            self.gui.csp.adjusted_image_path = os.path.join(self.gui.csp.working_directory, "adjusted_image.png")
        if round(self.gui.brightness_slider.value, 2) == 1 and round(self.gui.contrast_slider.value, 2) == 1 and not self.gui.auto_image_tuning.active:
            image = self.load_image(self.gui.csp.image_paths[self.gui.csp.image_id][self.gui.csp.channel_id])
            image.save(self.gui.csp.adjusted_image_path, format="PNG")
        else:
            image_data = base64.b64decode(self.gui.canvas.main_image.content.src_base64)
            buffer = BytesIO(image_data)
            image = Image.open(buffer)
            image.save(self.gui.csp.adjusted_image_path, format="PNG")


def auto_adjust(image_path,get_slice=-1):
    image = tifffile.imread(image_path)
    if image.ndim == 3:
        if not get_slice == -1:
            image = np.take(image, get_slice, axis=2)
        else:
            image = np.max(image, axis=2)
    image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
    normalized_image = cv2.normalize(image, None, alpha=0, beta=255, norm_type=cv2.NORM_MINMAX)

    _, buffer = cv2.imencode('.png', normalized_image)

    return base64.b64encode(buffer).decode('utf-8')


class AutoImageTuning:
    def __init__(self, gui):
        self.gui = gui
        self.active = False

    def pressed(self):
        if self.active:
            self.active = False
            self.gui.csp.config.set_auto_button(False)
            self.gui.auto_brightness_contrast.icon_color = ft.Colors.GREY_700
            if self.gui.csp.image_id is not None:
                self.gui.brightness_slider.disabled = False
                self.gui.contrast_slider.disabled = False
            self.gui.brightness_icon.color = None
            self.gui.contrast_icon.color = None
            self.gui.page.update()
            if self.gui.csp.image_id is not None:
                update_main_image(self.gui.csp.image_id, self.gui.csp.channel_id, self.gui, False)
        else:
            self.active = True
            self.gui.csp.config.set_auto_button(True)
            self.gui.auto_brightness_contrast.icon_color= ft.Colors.ORANGE_700
            if self.gui.csp.image_id is not None:
                self.gui.brightness_slider.disabled = True
                self.gui.contrast_slider.disabled = True
            self.gui.brightness_icon.color = ft.Colors.GREY_700
            self.gui.contrast_icon.color = ft.Colors.GREY_700
            self.gui.page.update()
            if self.gui.csp.image_id is not None:
                update_main_image(self.gui.csp.image_id, self.gui.csp.channel_id, self.gui, False)

    def update_main_image_auto(self,image_path):
        self.gui.canvas.main_image.content.src_base64 = auto_adjust(image_path)
        self.gui.canvas.main_image.update()




import asyncio
import multiprocessing
import os
import threading
import flet as ft

from cellsepi.backend.main_window.avg_diameter import AverageDiameter
from cellsepi.frontend.main_window.expert_mode.gui_builder import Builder
from cellsepi.frontend.main_window.expert_mode.gui_expert_environment import ExpertEnvironment
from cellsepi.frontend.main_window.gui_page_overlay import PageOverlay
from cellsepi.frontend.main_window.gui_segmentation import GUISegmentation
from cellsepi.frontend.main_window.gui_options import Options
from cellsepi.frontend.drawing_window.gui_drawing import open_qt_window
from cellsepi.frontend.main_window.gui_canvas import Canvas
from cellsepi.frontend.main_window.gui_config import GUIConfig
from cellsepi.frontend.main_window.gui_directory import DirectoryCard, copy_to_clipboard
from cellsepi.backend.main_window.cellsepi import CellSePi
from cellsepi.backend.main_window.mask import Mask
from cellsepi.frontend.main_window.gui_mask import error_banner, handle_image_switch_mask_on, handle_mask_update, reset_mask
from cellsepi.backend.main_window.image_tuning import ImageTuning, AutoImageTuning
from cellsepi.frontend.main_window.gui_training_environment import Training
from cellsepi.frontend.main_window.gui_page_overlay import PageOverlay
from cellsepi.frontend.main_window.expert_mode.expert_constants import ModuleType

class GUI:
    """
    Class GUI to handle the complete GUI and their attributes, also contains the CellSePi class and updates their attributes
    """
    def __init__(self,page: ft.Page):
        self.csp: CellSePi = CellSePi()
        self.page = page
        self.directory = DirectoryCard(self)
        self.switch_mask = ft.Switch(label="Mask", value=False)
        self.switch_mask.on_change = lambda e: self.update_view_mask()
        self.queue = multiprocessing.Queue()
        self.average_diameter = AverageDiameter(self)
        parent_conn, child_conn = multiprocessing.Pipe()
        self.parent_conn, self.child_conn = parent_conn, child_conn
        self.cancel_event = None
        self.closing_event = False
        self.training_event = None
        self.expert_running_event = None
        self.readout_event = None
        self.pipe_listener_running = True
        self.thread = threading.Thread(target=self.child_conn_listener, daemon=True)
        self.thread.start()
        self.page.window.prevent_close = True
        self.page.window.on_event = lambda e: self.handle_closing_event(e)
        self.process_drawing_window = self.start_drawing_window()
        self.drawing_button= ft.ElevatedButton(text="Drawing Tools", icon="brush_rounded",on_click=lambda e: self.set_queue_drawing_window(),disabled=True)
        self.page.window.width = 1428
        self.page.window.height = 800
        self.page.window.center()
        self.page.window.min_width = self.page.window.width
        self.page.window.min_height = self.page.window.height
        self.page.title = "CellSePi"
        self.canvas = Canvas()
        self.op = Options(self)
        self.ex_mode = ExpertEnvironment(self)
        gui_config = GUIConfig(self)
        self.gui_config = gui_config.create_profile_container()
        self.segmentation = GUISegmentation(self)
        seg_card,start_button,open_button,progress_bar,progress_bar_text,cancel_segmentation = self.segmentation.create_segmentation_card()
        self.cancel_segmentation = cancel_segmentation
        self.ready_to_start = False
        self.segmentation_card = seg_card
        self.open_button = open_button
        self.start_button = start_button
        self.progress_bar = progress_bar
        self.progress_bar_text = progress_bar_text
        self.mask=Mask(self.csp)
        self.image_tuning = ImageTuning(self)
        self.progress_ring = ft.ProgressRing(visible=False)
        self.closing_sheet = ft.Stack([
            ft.Column([ft.Container(ft.ProgressRing(),alignment=ft.alignment.center)],
                alignment=ft.MainAxisAlignment.CENTER,
            ),
        ])
        self.brightness_slider = ft.Slider(
            min=0, max=2.0, value=1.0, disabled= True,
            on_change=lambda e: asyncio.run(self.image_tuning.update_brightness_and_contrast_async())
        )
        self.contrast_slider = ft.Slider(
            min=0, max=2.0, value=1.0, disabled= True,
            on_change=lambda e: asyncio.run(self.image_tuning.update_brightness_and_contrast_async())
        )
        self.auto_image_tuning = AutoImageTuning(self)
        self.auto_brightness_contrast = ft.IconButton(icon=ft.Icons.AUTO_FIX_HIGH,icon_color=ft.Colors.GREY_700,style=ft.ButtonStyle(
                    shape=ft.RoundedRectangleBorder(radius=12),
                ),on_click=lambda e: self.auto_image_tuning.pressed(),tooltip="Auto brightness and contrast")
        self.brightness_icon = ft.Icon(name=ft.Icons.SUNNY,tooltip="Brightness")
        self.contrast_icon = ft.Icon(name=ft.Icons.CONTRAST,tooltip="Contrast")
        self.diameter_text = ft.Text("0.00", size=14, weight=ft.FontWeight.BOLD,tooltip="Copy to clipboard")
        self.diameter_display = ft.Container(
            content=ft.Row([ft.Icon(name=ft.Icons.STRAIGHTEN_ROUNDED, tooltip="Average diameter"), ft.GestureDetector(content=self.diameter_text,on_tap=lambda e: copy_to_clipboard(page=self.page,value=str(self.diameter_text.value),name="Average diameter"),on_enter=lambda e:self.on_enter_diameter(),on_exit=lambda e:self.on_exit_diameter()),]),
            border_radius=12,
            padding=8,
            opacity=0.5,
            visible=True,
        )
        self.training_environment=Training(self)
        self.ref_seg_environment = ft.Ref[ft.Column]()
        self.ref_training_environment = ft.Ref[ft.Column]()
        self.builder_environment = Builder(self)
        self.ref_builder_environment = ft.Ref[ft.Column]()
        self.ref_gallery_environment = ft.Ref[ft.Column]()
        if self.csp.config.get_auto_button():
            self.auto_image_tuning.pressed()

    def build(self):
        """
        Build up the main page of the GUI
        """
        self.page.add(
            ft.Column(
                [
                    ft.Row([
                            #LEFT COLUMN that handles all elements on the left side(canvas,switch_mask,segmentation)
                            ft.Column(
                        [
                                    self.canvas.canvas_card,
                                    ft.Row([self.switch_mask, self.drawing_button]),
                                    ft.Row([self.gui_config, ft.Column([ft.Card(content=ft.Container(content=ft.Column(
                                        [ft.Row([self.brightness_icon, ft.Container(self.brightness_slider, padding=-15)]),
                                         ft.Row([self.contrast_icon, ft.Container(self.contrast_slider, padding=-15)])]), padding=10)),
                                                                        ft.Row([ft.Card(content=self.auto_brightness_contrast),
                                                                                ft.Card(content=self.diameter_display)])])
                                            ]),
                                    self.segmentation_card
                                ],
                                expand=True,
                                alignment=ft.MainAxisAlignment.START,
                                visible=True,ref=self.ref_seg_environment
                            ),
                            ft.Column(
                        [
                                    self.training_environment.add_parameter_container(),
                                    self.training_environment.create_training_card()
                                ],
                                expand=True,
                                alignment=ft.MainAxisAlignment.START,
                                visible=False,ref=self.ref_training_environment
                            ),
                            ft.Column([self.builder_environment.builder_page_stack],expand=True,visible=False,ref=self.ref_builder_environment),
                            #RIGHT COLUMN that handles gallery and directory_card
                            ft.Column(
                                [
                                    self.directory,
                                    ft.Card(
                                        content=ft.Stack([ft.Container(self.progress_ring,alignment=ft.alignment.center),ft.Container(self.directory.image_gallery,padding=20)]),
                                        expand=True
                                    ),
                                ],
                                expand=True,ref=self.ref_gallery_environment
                            ),
                            ft.Column([self.op, self.training_environment,self.ex_mode]),
                        ],
                        alignment=ft.MainAxisAlignment.SPACE_BETWEEN,
                        expand=True,
                    ),
                ],
                expand=True
            ),
        )
        #set the colors for the review module from the config file
        ModuleType.REVIEW.value.mask_color = self.csp.config.get_mask_color()
        ModuleType.REVIEW.value.outline_color = self.csp.config.get_outline_color()
        ModuleType.REVIEW.value.update_class()

    def update_view_mask(self):
        """
        Method that controls what happened when switch is on/off
        """
        if self.csp.image_id is None:
            error_banner(self,"No image selected!")
            self.switch_mask.value=False
        else:
            handle_image_switch_mask_on(self)

    def start_drawing_window(self):
        """
        Start the drawing window in multiprocessing.
        """
        process = multiprocessing.Process(target=open_qt_window,
                                args=(self.queue,self.child_conn))
        process.start()
        return process

    def set_queue_drawing_window(self):
        """
        Sets queue for drawing window with the current selected image and mask.
        """
        self.image_tuning.save_current_main_image()
        if self.process_drawing_window is None or not self.process_drawing_window.is_alive(): #make sure that the process is running before putting new image in the queue
            if self.process_drawing_window is not None:
                try:
                    self.process_drawing_window.terminate()
                    self.process_drawing_window.join()
                except Exception as e:
                    self.page.open(ft.SnackBar(ft.Text(f"Error while terminating process: {e}")))
            self.queue = multiprocessing.Queue()
            parent_conn, child_conn = multiprocessing.Pipe()
            self.parent_conn, self.child_conn = parent_conn, child_conn
            self.process_drawing_window = self.start_drawing_window()
        self.csp.window_image_id = self.csp.image_id
        self.csp.window_bf_channel = self.csp.config.get_bf_channel()
        self.csp.window_channel_id = self.csp.channel_id

        if self.csp.window_bf_channel in self.csp.image_paths[self.csp.image_id]:#check if the bf has an image
            image_path = self.csp.image_paths[self.csp.image_id][self.csp.window_bf_channel]
            directory, filename = os.path.split(image_path)
            name, _ = os.path.splitext(filename)
            mask_file_name = f"{name}{self.csp.current_mask_suffix}.npy"
            self.csp.window_mask_path= os.path.join(directory, mask_file_name)
            self.queue.put((self.csp.config.get_mask_color(), self.csp.config.get_outline_color(),self.csp.color_opacity, self.csp.window_bf_channel, self.csp.mask_paths, self.csp.window_image_id, self.csp.adjusted_image_path, self.csp.window_mask_path,self.csp.window_channel_id,self.csp.current_channel_prefix))
        else:
            self.page.open(ft.SnackBar(
                ft.Text(f"Selected bright-field channel {self.csp.window_bf_channel} has no image!")))
            self.page.update()

    def handle_closing_event(self, e,saved_checked:bool=False):
        """
        Handle the closing event of Flet GUI.
        """
        if e.data == "close" and not self.closing_event:
            if not self.builder_environment.pipeline_storage.check_saved() and not saved_checked:
                def cancel_dialog(a):
                    cupertino_alert_dialog.open = False
                    a.control.page.update()

                def ok_dialog(a,gui):
                    cupertino_alert_dialog.open = False
                    a.control.page.update()
                    gui.handle_closing_event(e,True)

                cupertino_alert_dialog = ft.CupertinoAlertDialog(
                    title=ft.Text("Expert Mode:\nUnsaved Changes"),
                    content=ft.Text("Closing CellSePi will discard any unsaved changes to the currently opened pipeline."),
                    actions=[
                        ft.CupertinoDialogAction(
                            "Cancel", is_default_action=True, on_click=cancel_dialog
                        ),
                        ft.CupertinoDialogAction(text="Ok", is_destructive_action=True, on_click=lambda a: ok_dialog(a,self)),
                    ],
                )
                self.page.overlay.append(cupertino_alert_dialog)
                cupertino_alert_dialog.open = True
                self.page.update()
                return

            self.closing_event = True
            overlay = PageOverlay(self.page,content=self.closing_sheet,modal=True)
            overlay.open()
            if self.csp.segmentation_running:
                self.cancel_event = multiprocessing.Event()
                self.cancel_segmentation()
                self.cancel_event.wait()
            if self.csp.training_running:
                self.training_event = multiprocessing.Event()
                self.training_event.wait()
            if self.csp.readout_running:
                self.readout_event = multiprocessing.Event()
                self.readout_event.wait()
            if self.builder_environment.pipeline_gui.pipeline.running:
                self.builder_environment.cancel()
                self.expert_running_event = multiprocessing.Event()
                self.builder_environment.pipeline_running_event = self.expert_running_event
                self.expert_running_event.wait()
            self.pipe_listener_running = False
            self.queue.put("close")
            if self.process_drawing_window is not None and self.process_drawing_window.is_alive():
                self.process_drawing_window.join()
            self.child_conn.send("close")

            if self.thread is not None and self.thread.is_alive():
                self.thread.join()
            self.child_conn.close()
            self.parent_conn.close()
            self.page.window.prevent_close = False
            self.page.window.on_event = None
            self.page.update()
            self.page.window.close()

    def child_conn_listener(self):
        """
        Listener for the child connection.
        """
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        async def pipe_listener():
            while self.pipe_listener_running:
                data = await asyncio.to_thread(self.parent_conn.recv)
                split_data = data.split(".")
                if data == "close":
                    break
                elif split_data[0] == "new_mask":
                    if self.csp.window_image_id not in self.csp.mask_paths:
                        self.csp.mask_paths[self.csp.window_image_id] = {}
                    self.csp.mask_paths[self.csp.window_image_id][self.csp.window_bf_channel] = self.csp.window_mask_path
                    self.directory.update_mask_check(split_data[1])
                    self.page.run_task(self.directory.check_masks)
                elif split_data[0] == "ready":
                    pass
                else:
                    if self.csp.window_image_id == self.csp.image_id and self.csp.window_bf_channel == self.csp.config.get_bf_channel() and self.switch_mask.value:
                        handle_mask_update(self)
                        self.page.update()
                    else:
                        reset_mask(self,self.csp.window_image_id,self.csp.window_bf_channel)
                    self.diameter_text.value = self.average_diameter.get_avg_diameter()
                    self.diameter_display.visible = True
                    self.diameter_display.update()
        try:
            loop.run_until_complete(pipe_listener())
        finally:
            loop.stop()
            loop.close()

    def on_enter_diameter(self):
        self.diameter_text.color = ft.Colors.BLUE_400
        self.diameter_text.update()
    def on_exit_diameter(self):
        self.diameter_text.color = None
        self.diameter_text.update()



from setuptools import setup, find_packages
import os

this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, "README.md"), encoding="utf-8") as f:
    long_description = f.read()


setup(
    name="cellsepi",
    version="1.1.4",
    license="Apache License 2.0",
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    install_requires=[
        "numpy==1.26.4", "pillow", "pandas", "openpyxl", "cellpose==3.1.1.1",
        "flet==0.28.3","flet-cli==0.28.3", "flet-desktop==0.28.3", "flet-runtime==0.24.1", "bioio==1.2.0",
        "numba==0.61.0", "matplotlib", "pytest", "pyqt5", "flet_contrib", "flet_core==0.24.1",
        "bioio-lif","torchvision == 0.17.2","jsonschema == 4.25.1","flet-extended-interactive-viewer==0.1.2",
        "scikit-image==0.25.2","big-fish == 0.6.2"
    ],
    python_requires=">=3.8",
    author="Jenna Ahlvers, Santosh Chhetri Thapa, Nike Dratt, Pascal Heß, Florian Hock",
    url="https://github.com/PraiseTheDarkFlo/CellSePi",
    description="Segmentation of microscopy images and data analysis pipeline with a graphical user interface, powered by Cellpose.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    entry_points={
        "console_scripts": [
            "cellsepi = cellsepi.main:main",
        ],
    },
    package_data={
        "cellsepi": ["models/*", "models/**/*"]
    },
    include_package_data=True
)




from cellsepi.backend.main_window.data_util import load_directory, ReturnTypePath
from cellsepi.backend.main_window.expert_mode.listener import ProgressEvent
from cellsepi.backend.main_window.expert_mode.module import *
from cellsepi.backend.main_window.expert_mode.pipeline import PipelineRunningException
from cellsepi.backend.main_window.images import BatchImageSegmentation
from cellsepi.frontend.main_window.gui_directory import DirectoryCard


class ImageSegmentationModule(Module, ABC):
    _gui_config = ModuleGuiConfig("ImageSegmentation",Categories.SEGMENTATION,"This module handles the segmentation of cells for each series on the given segmentation_channel with the provided model in model_path.")
    def __init__(self, module_id: str) -> None:
        super().__init__(module_id)
        self.inputs = {
            "image_paths": InputPort("image_paths", dict),
            "mask_paths": InputPort("mask_paths", dict,opt=True),
        }
        self.outputs = {
            "mask_paths": OutputPort("mask_paths", dict),
        }
        self.user_model_path: FilePath = FilePath()
        self.user_segmentation_channel: str = "2"
        self.user_diameter: float = 125.0
        self.user_mask_suffix: str = "_seg"

    def run(self):
        if self.inputs["mask_paths"].data is None:
            self.inputs["mask_paths"].data = {}
        try:
            BatchImageSegmentation(segmentation_channel=self.user_segmentation_channel,diameter=self.user_diameter,suffix=self.user_mask_suffix).run(self.event_manager,self.inputs["image_paths"].data,self.inputs["mask_paths"].data,self.user_model_path.path)
        except:
            raise PipelineRunningException("Segmentation Error", "Incompatible file for the segmentation model.")

        self.outputs["mask_paths"].data = self.inputs["mask_paths"].data




from cellsepi.backend.main_window.config_file import ConfigFile


class CellSePi:
    def __init__(self):
        super().__init__()
        self.config: ConfigFile = ConfigFile()
        self.segmentation_running = False
        self.segmentation_thread = None
        self.training_running = False
        self.model_path = None
        self.re_train_model_path=None
        self.readout_running = False
        self.readout_thread = None
        self.readout_path = None
        self.linux_or_3d = False

        self.image_id = None
        self.channel_id = None
        self.window_image_id = None
        self.window_bf_channel = None
        self.window_channel_id = None
        self.window_mask_path = None
        self.current_channel_prefix = None
        self.current_mask_suffix = None
        self.window_mask_path = None
        self.color_opacity = 128


        self.readout = None

        self.adjusted_image_path = None
        self.image_paths = None #[image_id, different images sorted by channel]
        self.linux_images = None #[image_id][channel_id] = base64 png image
        self.mask_paths = None
        self.working_directory = None




import copy

from cellsepi.backend.main_window.expert_mode.module import Module
from typing import List

IMMUTABLES = (int, float, str, bool, type(None))

def copy_data(value):
    """
    To take care that no references are transferred between the modules.
    """
    if isinstance(value, IMMUTABLES):
        return value

    return copy.deepcopy(value)

class Pipe:
    """
    Transfers the data from one module to another.
    """
    def __init__(self,source_module: Module,target_module: Module, ports: List[str]):
        """
        Raises:
            ValueError: If the source and target modules do not different.
        """
        if source_module.module_id == target_module.module_id:
            raise ValueError(f"Source and target modules must be different!")
        if ports is None or len(ports) == 0:
            raise ValueError(f"Ports must be non empty!")
        self.source_module = source_module
        self.target_module = target_module
        self.ports = ports

    def run(self):
        """
        Transfers the data from source modules outputs to target modules inputs.
        Raises:
            KeyError: If the source and target modules not have the ports in the outputs/inputs.
            TypeError: If the types of the port on the source and target modules do not match.
        """
        for port in self.ports:
            if port not in self.source_module.outputs:
                raise KeyError(f"'{port}' is not in the outputs of Module '{self.source_module.module_id}'")
            if port not in self.target_module.inputs:
                raise KeyError(f"'{port}' is not in the inputs of Module '{self.target_module.module_id}'")
            out_port = self.source_module.outputs[port]
            in_port = self.target_module.inputs[port]

            if out_port.data_type != in_port.data_type:
                raise TypeError(
                    f"Type mismatch on port '{port}', source_module provided {out_port.data_type.__name__}, output_module provided {in_port.data_type.__name__}!")

            in_port.data = copy_data(out_port.data)

    def __str__(self):
        return f"source: '{self.source_module.module_id}', target: '{self.target_module.module_id}', ports: {self.ports}"

    def to_dict(self):
        """
        Translates the pipe into a dictionary.
        """
        return {
            "source": self.source_module.module_id,
            "target": self.target_module.module_id,
            "ports": self.ports,
        }


import multiprocessing

import flet as ft
from cellsepi.frontend.main_window.gui import GUI

def main():
    multiprocessing.set_start_method("spawn")
    ft.app(target=async_main, view=ft.FLET_APP)

async def async_main(page: ft.Page):
    gui = GUI(page)
    gui.build()
